'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.transpile = transpile;
exports.compile = compile;

var _astTypes = require('ast-types');

var _recast = require('recast');

var _recast2 = _interopRequireDefault(_recast);

var _coffeeScript = require('coffee-script');

var _scope = require('coffee-script/lib/coffee-script/scope');

var _findWhere = require('lodash/collection/findWhere');

var _findWhere2 = _interopRequireDefault(_findWhere);

var _last = require('lodash/array/last');

var _last2 = _interopRequireDefault(_last);

var _flatten = require('lodash/array/flatten');

var _flatten2 = _interopRequireDefault(_flatten);

var _pick = require('lodash/object/pick');

var _pick2 = _interopRequireDefault(_pick);

var _values = require('lodash/object/values');

var _values2 = _interopRequireDefault(_values);

var _findIndex = require('lodash/array/findIndex');

var _findIndex2 = _interopRequireDefault(_findIndex);

var _get = require('lodash/object/get');

var _get2 = _interopRequireDefault(_get);

var _compose = require('lodash/function/compose');

var _compose2 = _interopRequireDefault(_compose);

var _isArray = require('lodash/lang/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _any = require('lodash/collection/any');

var _any2 = _interopRequireDefault(_any);

var _jscodeshift = require('jscodeshift');

var _jscodeshift2 = _interopRequireDefault(_jscodeshift);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

// regexes taken from coffeescript parser
var IS_NUMBER = /^[+-]?(?:0x[\da-f]+|\d*\.?\d+(?:e[+-]?\d+)?)$/i;
var IS_STRING = /^['"]/;
var IS_REGEX = /^\//;

var STRING_INSIDE_QUOTES = /^['"](.*)['"]$/;

function pluck(obj, keys) {
  return (0, _values2.default)((0, _pick2.default)(obj, keys));
}

function mapBoolean(node) {
  if (node.base.val === 'true') {
    return _astTypes.builders.literal(true);
  } else if (node.base.val === 'false') {
    return _astTypes.builders.literal(false);
  }

  throw new Error('can\'t convert node of type: ' + node.constructor.name + ' to boolean - not recognized');
}

function stringToRegex(inputstring) {
  var match = inputstring.match(new RegExp('^/(.*?)/([gimy]*)$'));
  return new RegExp(match[1], match[2]);
}

function mapMemberProperties(properties, meta) {
  var restProperties = properties.slice(0, properties.length - 1);
  var isIndex = properties[properties.length - 1].constructor.name === 'Index';
  var right = mapExpression(properties[properties.length - 1], meta);
  var isComputed = right.type === 'Literal' || isIndex;
  var left = undefined;

  if (restProperties.length === 1) {
    left = mapExpression(restProperties[0], meta);
  } else {
    left = mapMemberProperties(restProperties, meta);
  }

  return _astTypes.builders.memberExpression(left, right, isComputed);
}

function mapMemberExpression(node, meta) {
  if ((0, _findIndex2.default)(node.base.properties, { soak: true }) > -1) {
    return fallback(node, meta);
  }
  return mapMemberProperties([node.base].concat(_toConsumableArray(node.properties)), meta);
}

function mapLiteral(node) {
  var value = undefined;
  value = node.base.value;

  if (value === 'NaN') {
    return _astTypes.builders.literal(NaN);
  } else if (IS_STRING.test(value)) {
    return _astTypes.builders.literal(value.match(STRING_INSIDE_QUOTES)[1]);
  } else if (IS_NUMBER.test(value)) {
    return _astTypes.builders.literal(Number(value));
  } else if (IS_REGEX.test(value)) {
    return _astTypes.builders.literal(stringToRegex(value));
  }

  return _astTypes.builders.identifier(value);
}

function mapKey(node) {
  var type = node.base.constructor.name;
  if (type === 'Literal') {
    return _astTypes.builders.identifier(node.base.value);
  }
}

function mapObjectExpression(node, meta) {
  return _astTypes.builders.objectExpression(node.base.properties.map(function (property) {
    return _astTypes.builders.property('init', mapExpression(property.variable || property.base, meta), mapExpression(property.value || property.base, meta));
  }));
}

function mapArrayExpression(node, meta) {
  return _astTypes.builders.arrayExpression(node.objects.map(function (expr) {
    return mapExpression(expr, meta);
  }));
}

function mapRange(node, meta) {
  var compiledRange = _recast2.default.parse(_recast2.default.prettyPrint(_recast2.default.parse(node.compile(meta)))).program.body[0];
  return compiledRange.expression;
}

function mapSlice(node, meta) {
  return _astTypes.builders.callExpression(_astTypes.builders.identifier('slice'), [mapExpression(node.range.from, meta), mapExpression(node.range.to, meta)]);
}

function mapValue(node, meta) {
  var type = node.base.constructor.name;

  if (type === 'Literal') {
    return mapLiteral(node, meta);
  } else if (type === 'Range') {
    return mapRange(node, meta);
  } else if (type === 'Undefined') {
    return _astTypes.builders.identifier('undefined');
  } else if (type === 'Null') {
    return _astTypes.builders.identifier('null');
  } else if (type === 'Call') {
    return mapCall(node.base, meta);
  } else if (type === 'Bool') {
    return mapBoolean(node, meta);
  } else if (type === 'Arr' && meta.left === true) {
    return mapArrayPattern(node.base, meta);
  } else if (type === 'Obj' && meta.left === true) {
    return mapObjectPattern(node.base.properties, meta);
  } else if (type === 'Arr') {
    return mapArrayExpression(node.base, meta);
  } else if (type === 'Obj') {
    return mapObjectExpression(node, meta);
  } else if (type === 'Parens') {
    return _astTypes.builders.sequenceExpression(node.base.body.expressions.map(function (expr) {
      return mapExpression(expr, meta);
    }));
  }

  throw new Error('can\'t convert node of type: ' + type + ' to value - not recognized');
}

function mapOp(node, meta) {
  var operator = node.operator;

  if (node.operator === '%%' && node.second) {
    var op = _astTypes.builders.binaryExpression('%', _astTypes.builders.parenthesizedExpression(_astTypes.builders.binaryExpression('+', _astTypes.builders.binaryExpression('%', mapExpression(node.first, meta), mapExpression(node.second, meta)), mapExpression(node.second, meta))), mapExpression(node.second, meta));
    return op;
  }

  if (node.operator === '++' || node.operator === '--') {
    return _astTypes.builders.updateExpression(node.operator, mapExpression(node.first, meta), !node.flip);
  }

  if (!node.second) {
    return _astTypes.builders.unaryExpression(node.operator, mapExpression(node.first, meta));
  }

  if (operator === '||' || operator === '&&') {
    return _astTypes.builders.logicalExpression(node.operator, mapExpression(node.first, meta), mapExpression(node.second, meta));
  }

  return _astTypes.builders.binaryExpression(node.operator, mapExpression(node.first, meta), mapExpression(node.second, meta));
}

function mapArguments(args, meta) {
  return args.map(function (arg) {
    var type = undefined;
    if (arg.name && arg.name.constructor) {
      type = arg.name.constructor.name;
    }
    if (type === 'Arr') {
      return mapArrayPattern(arg.name, meta);
    } else if (type === 'Obj') {
      return mapObjectPattern(arg.name.properties, meta);
    }

    return mapExpression(arg, meta);
  });
}

function mapCall(node, meta) {
  var left = undefined;
  var superMethodName = meta.superMethodName;

  if (node.soak === true) {
    return _recast2.default.parse(node.compile(meta)).program.body[0].expression;
  } else if (node.isSuper === true && superMethodName === 'constructor') {
    left = _astTypes.builders.identifier('super');
  } else if (node.isSuper === true && superMethodName !== undefined) {
    left = _astTypes.builders.memberExpression(_astTypes.builders.identifier('super'), _astTypes.builders.identifier(superMethodName));
  } else {
    left = mapExpression(node.variable, meta);
  }

  return _astTypes.builders.callExpression(left, mapArguments(node.args, meta));
}

function mapAssignment(node, meta) {
  return _astTypes.builders.expressionStatement(mapExpression(node, meta));
}

function mapClassProperty(node, meta) {
  return _astTypes.builders.classProperty(mapExpression(node.variable, meta), mapExpression(node.value, meta), null);
}

function mapClassBodyElement(node, meta) {
  var superMethodName = node.variable.base.value;
  var elementType = 'method';
  var isStatic = false;

  if (node.variable.this === true) {
    isStatic = true;
    node.variable = (0, _get2.default)(node, 'variable.properties[0].name');
  }

  if (node.constructor.name === 'Assign' && node.value && node.value.constructor.name !== 'Code') {
    return mapClassProperty(node, meta);
  }

  if (superMethodName === 'constructor') {
    elementType = 'constructor';
  }

  var _meta = _extends({}, meta, { superMethodName: superMethodName });

  return _astTypes.builders.methodDefinition(elementType, mapExpression(node.variable, _meta), mapExpression(node.value, _meta), isStatic);
}

function getBoundMethodNames(classElements, meta) {
  return (0, _flatten2.default)(classElements.filter(function (el) {
    return el.base && el.base.properties;
  }).map(function (el) {
    return el.base.properties;
  })).filter(function (el) {
    return (0, _get2.default)(el, 'variable.this') !== true && el.value.constructor.name === 'Code' && el.value.bound === true;
  }).map(function (el) {
    return mapExpression(el.variable, meta);
  });
}

function unbindMethods(classElements) {
  return classElements.map(function (el) {
    if (el.value.constructor.name === 'Code') {
      el.value.bound = false;
    }
    return el;
  });
}

function mapStaticClassProperty(node, meta) {
  return _astTypes.builders.classProperty(mapExpression(node.variable.properties[0], meta), mapExpression(node.value, meta), null, true);
}

function mapClassExpressions(expressions, meta) {
  return expressions.reduce(function (arr, expr) {
    var type = expr.constructor.name;
    var classElements = [];
    if (type === 'Assign') {
      if (expr.variable && expr.variable.this === true) {
        return arr.concat([mapStaticClassProperty(expr, meta)]);
      }
    } else if (type === 'Value') {
      classElements = expr.base.properties;
      classElements = unbindMethods(classElements);
      classElements = classElements.map(function (el) {
        return mapClassBodyElement(el, meta);
      });
      return arr.concat(classElements);
    }
    return arr;
  }, []);
}

function mapClassBody(node, meta) {
  var expressions = node.expressions;

  var boundMethods = getBoundMethodNames(expressions, meta);
  var classElements = mapClassExpressions(expressions, meta);
  var constructor = (0, _findWhere2.default)(classElements, { kind: 'constructor' });

  if (boundMethods.length > 0) {
    if (constructor === undefined) {
      // create an empty constructor if there isn't one yet
      constructor = _astTypes.builders.methodDefinition('constructor', _astTypes.builders.identifier('constructor'), _astTypes.builders.functionExpression(null, [], _astTypes.builders.blockStatement([])));
      classElements.unshift(constructor);
    }

    // bind all the bound methods to the class
    constructor.value.body.body = constructor.value.body.body.concat(boundMethods.map(function (identifier) {
      return _astTypes.builders.expressionStatement(_astTypes.builders.callExpression(_astTypes.builders.memberExpression(_astTypes.builders.memberExpression(_astTypes.builders.thisExpression(), identifier), _astTypes.builders.identifier('bind')), [_astTypes.builders.thisExpression()]));
    }));
  }

  return _astTypes.builders.classBody(classElements);
}

function mapClassExpression(node, meta) {
  // if this is an anonymous class expression fallback
  // to the cs compiler
  if (node.variable === undefined && node.parent === undefined && node.body.expressions.length < 1) {
    return fallback(node, meta);
  }

  var parent = null;

  if (node.parent !== undefined && node.parent !== null) {
    parent = mapExpression(node.parent, meta);
  }

  return _astTypes.builders.classExpression(mapExpression(node.variable, meta), mapClassBody(node.body, meta), parent);
}

function mapClassDeclaration(node, meta) {
  var parent = null;

  if ((0, _get2.default)(node, 'variable.properties.length') > 0) {
    return _astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', mapExpression(node.variable, meta), mapClassExpression(_extends({}, node, { variable: (0, _last2.default)(node.variable.properties) }))));
  }

  if (node.parent !== undefined && node.parent !== null) {
    parent = mapExpression(node.parent, meta);
  }

  return _astTypes.builders.classDeclaration(mapExpression(node.variable, meta), mapClassBody(node.body, meta), parent);
}

function mapElseBlock(node, meta) {
  var type = node.constructor.name;

  if (type === 'If') {
    var conditional = mapIfStatement(node, meta);
    if (_astTypes.namedTypes.IfStatement.check(conditional)) {
      return conditional;
    }
    return _astTypes.builders.blockStatement([conditional]);
  } else if (type === 'Block') {
    return mapBlockStatement(node, meta);
  }

  return mapBlockStatement({ expressions: [node] }, meta);
}

function mapElseExpression(node, meta) {
  var type = node.constructor.name;

  if (type === 'If') {
    return mapConditionalExpression(node, meta);
  } else if (type === 'Block') {
    return mapExpression(node.expressions[0], meta);
  }

  return mapExpression(node, meta);
}

function mapConditionalExpression(node, meta) {
  var alternate = _astTypes.builders.identifier('undefined');

  if (node.elseBody) {
    alternate = mapElseExpression(node.elseBody.expressions[0], meta);
  }

  return _astTypes.builders.conditionalExpression(mapExpression(node.condition, meta), mapExpression(node.body.expressions[0], meta), alternate);
}

function mapTryExpression(node, meta) {
  var tryBlock = mapTryCatchBlock(node, meta);
  tryBlock.block = addReturnStatementToBlock(tryBlock.block);
  return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([tryBlock])), []);
}

function mapIfStatement(node, meta) {
  var alternate = null;
  if (node.elseBody) {
    alternate = mapElseBlock(node.elseBody.expressions[0], meta);
  }
  return _astTypes.builders.ifStatement(mapExpression(node.condition, meta), mapBlockStatement(node.body, meta), alternate);
}

function isStatement(expr) {
  var type = expr.constructor.name;
  switch (type) {
    case 'Literal':
      if (expr.value === 'break') {
        return true;
      }
      return false;
    case 'Throw':
    case 'For':
    case 'While':
    case 'Return':
    case 'If':
    case 'Break':
      return true;
    default:
      return false;
  }
}

function mapConditionalStatement(node, meta) {
  // If the conditional has more than one test
  // or more than one expression in either block we
  // create an if statement otherwise we use a conditional
  // expression (this is all just for readability)
  if (node.elseBody && node.elseBody.expressions.length > 1 || node.body && node.body.expressions.length > 1 || node.body && (0, _any2.default)(node.body.expressions, function (expr) {
    return isStatement(expr);
  }) || node.elseBody && (0, _any2.default)(node.elseBody.expressions, function (expr) {
    return isStatement(expr);
  })) {
    return mapIfStatement(node, meta);
  }

  return _astTypes.builders.expressionStatement(mapConditionalExpression(node, meta));
}

function mapTryCatchBlock(node, meta) {
  var recovery = undefined;
  var errorVar = undefined;
  var finalize = null;

  if (node.ensure) {
    finalize = mapBlockStatement(node.ensure, meta);
  }

  if (node.recovery) {
    recovery = mapBlockStatement(node.recovery, meta);
  } else {
    recovery = _astTypes.builders.blockStatement([]);
  }

  if (node.errorVariable) {
    errorVar = mapLiteral({ base: node.errorVariable }, meta);
  } else {
    errorVar = _astTypes.builders.identifier('undefined');
  }

  return _astTypes.builders.tryStatement(mapBlockStatement(node.attempt, meta), _astTypes.builders.catchClause(errorVar, null, recovery), finalize);
}

function mapReturnStatement(node, meta) {
  return _astTypes.builders.returnStatement(node.expression ? mapExpression(node.expression, meta) : null);
}

function mapStatement(node, meta) {
  var type = node.constructor.name;

  if (type === 'While') {
    return mapWhileLoop(node, meta);
  } else if (type === 'Assign') {
    return mapAssignment(node, meta);
  } else if (type === 'Return') {
    return mapReturnStatement(node, meta);
  } else if (type === 'Throw') {
    return mapThrowStatement(node, meta);
  } else if (type === 'Comment') {
    return _astTypes.builders.emptyStatement();
  } else if (type === 'For') {
    return mapForStatement(node, meta);
  } else if (type === 'Class') {
    return mapClassDeclaration(node, meta);
  } else if (type === 'Switch') {
    return addBreakStatementsToSwitch(mapSwitchStatement(node, meta));
  } else if (type === 'If') {
    return mapConditionalStatement(node, meta);
  } else if (type === 'Try') {
    return mapTryCatchBlock(node, meta);
  }

  return _astTypes.builders.expressionStatement(mapExpression(node, meta));
}

function mapBlockStatements(node, meta) {
  return node.expressions.map(function (expr) {
    return mapStatement(expr, meta);
  });
}

function mapBlockStatement(node, meta) {
  var factory = arguments.length <= 2 || arguments[2] === undefined ? _astTypes.builders.blockStatement : arguments[2];

  var block = factory(mapBlockStatements(node, meta));
  return block;
}

function mapInArrayExpression(node, meta) {
  return _astTypes.builders.memberExpression(mapExpression(node.array, meta), _astTypes.builders.callExpression(_astTypes.builders.identifier('includes'), [mapExpression(node.object, meta)]));
}

function extractAssignStatementsByArguments(nodes) {
  return nodes.map(function (node) {
    return node.type === 'AssignmentExpression' ? node.left : node;
  }).filter(function (node) {
    return node.type === 'MemberExpression' && (node.object.type === 'ThisExpression' || node.object.name === 'this');
  }).map(function (node) {
    return _astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', node, node.property));
  });
}

function normalizeArguments(nodes) {
  return nodes.map(function (node) {
    if (node.type === 'AssignmentExpression' && node.left.type === 'MemberExpression') {
      return _astTypes.builders.assignmentExpression(node.operator, node.left.property, node.right);
    }
    if (node.type === 'MemberExpression' && (node.object.type === 'ThisExpression' || node.object.name === 'this')) {
      return { type: 'Identifier', name: node.property.name };
    }
    return node;
  });
}

// index can be an expression in this case
function mapArgumentWithExpansion(node, meta, arg) {
  var expr = mapExpression(node.name, meta);
  var statements = [];

  if (node.name.base && node.name.base.value === 'this') {
    statements.push(_astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', expr, arg)));
  } else {
    statements.push(_astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(expr, arg)]));
  }

  if (node.value !== undefined && node.value !== null) {
    statements.push(_astTypes.builders.ifStatement(_astTypes.builders.binaryExpression('===', arg, _astTypes.builders.identifier('undefined')), _astTypes.builders.expressionStatement(_astTypes.builders.assignmentExpression('=', expr, mapExpression(node.value, meta)))));
  }
  return statements;
}

function mapSplatArgument(headLength, tailCount) {
  return _astTypes.builders.memberExpression(_astTypes.builders.identifier('arguments'), _astTypes.builders.callExpression(_astTypes.builders.identifier('slice'), [_astTypes.builders.literal(headLength), _astTypes.builders.memberExpression(_astTypes.builders.identifier('arguments'), _astTypes.builders.binaryExpression('-', _astTypes.builders.memberExpression(_astTypes.builders.identifier('arguments'), _astTypes.builders.identifier('length')), _astTypes.builders.literal(tailCount)), true)]));
}

function mapArgumentsWithExpansion(nodes, meta) {
  // In coffeescript you can have arguments that are
  // positioned at the end like this: fn = (begin, middle..., end) ->
  // The bit in the middle is of type Splat or Expansion. Expansion
  // purely exists for defining such a 'last' Argument. There can be
  // one or more last arguments. There can not be more than one
  // argument of type Splat or Expansion however.
  // Last arguments behave as normal

  // Initiate statements variable, we'll fill this up.
  var statements = [];

  // find expansion index
  var expansionIndex = (0, _findIndex2.default)(nodes, function (node) {
    return node.constructor.name === 'Expansion' || node.splat === true;
  });
  var isSplat = (0, _any2.default)(nodes, function (node) {
    return node.splat === true;
  });

  // separate head[] from tail[], omit the index.
  var head = nodes.slice(0, expansionIndex);
  var tail = nodes.slice(expansionIndex + 1, nodes.length);

  if (isSplat === true) {
    // if the expansion is a splat we'll add it to head
    // so it gets processed
    head.push(nodes[expansionIndex]);
  }

  // loop over head and tail and create assignment expression
  // statements

  // build head[] statements
  head.forEach(function (node, index) {
    var argument = undefined;
    if (node.splat === true) {
      argument = mapSplatArgument(head.length, tail.length);
    } else {
      argument = _astTypes.builders.memberExpression(_astTypes.builders.identifier('arguments'), _astTypes.builders.literal(index), true);
    }

    statements.push.apply(statements, mapArgumentWithExpansion(node, meta, argument));
  });

  // build tail[] statements
  tail.reverse().forEach(function (node, index) {
    var argument = _astTypes.builders.memberExpression(_astTypes.builders.identifier('arguments'), _astTypes.builders.binaryExpression('-', _astTypes.builders.memberExpression(_astTypes.builders.identifier('arguments'), _astTypes.builders.identifier('length')), _astTypes.builders.literal(index + 1)), true);

    statements.push.apply(statements, mapArgumentWithExpansion(node, meta, argument));
  });

  return statements;
}

function transformToExpression(_node) {
  var node = _node;

  if (node.expression !== undefined) {
    return node.expression;
  }

  if (node.type === 'IfStatement') {
    node = addReturnStatementToIfBlocks(node);
  } else if (node.tpye === 'SwitchStatement') {
    node = node;
  }

  return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([node])), []);
}

function lastReturnStatement() {
  var nodeList = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];

  if (nodeList.length > 0) {
    var lastIndex = nodeList.length - 1;

    if (nodeList[lastIndex].type === 'IfStatement') {
      nodeList[lastIndex] = addReturnStatementToIfBlocks(nodeList[lastIndex]);
    } else {
      nodeList[lastIndex] = _astTypes.builders.returnStatement(transformToExpression(nodeList[nodeList.length - 1]));
    }
  }
  return nodeList;
}

function lastBreakStatement() {
  var nodeList = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];

  if (nodeList.length > 0) {
    nodeList.push(_astTypes.builders.breakStatement());
  }
  return nodeList;
}

function addReturnStatementToIfBlocks(node) {
  node.consequent = addReturnStatementToBlock(node.consequent);
  if (_astTypes.namedTypes.IfStatement.check(node.alternate)) {
    node.alternate = addReturnStatementToIfBlocks(node.alternate);
  } else if (_astTypes.namedTypes.ExpressionStatement.check(node.alternate)) {
    node.alternate = _astTypes.builders.returnStatement(node.alternate.expression);
  } else if (node.alternate) {
    node.alternate = addReturnStatementToBlock(node.alternate);
  }
  return node;
}

function addReturnStatementToBlock(node) {
  var hasReturnStatement = (0, _findIndex2.default)(node.body, { type: 'ReturnStatement' }) === node.body.length - 1;

  if (hasReturnStatement) {
    return node;
  }
  node.body = lastReturnStatement(node.body);
  return node;
}

function mapFunction(node, meta) {
  // Function {
  //   params: [],
  //   body: [statements],
  //   bound: Boolean
  // }
  var args = [];

  // setupStatements will be appended at the top of the function
  // block. It's used to add behaviour that would be impossible to
  // map 1 to 1 from coffeescript
  var setupStatements = [];

  // Expansions are coffee-script-only behaviour. We'll need to do
  // some plumbing to map Expansions to JavaScript, also we don't
  // want to rely on the coffeescript parser for this as the output
  // is quite weird/ugly
  var hasExpansion = (0, _any2.default)(node.params, function (param, index) {
    return(
      // if the last argument isn't a splat or expansion we needn't worry
      // To spell it out the logic here is:
      // If this isn't the last argument, and
      // the argument is either an expansion or a splat
      // then we return true
      index < node.params.length - 1 && (param.constructor.name === 'Expansion' || param.splat === true)
    );
  });

  if (hasExpansion === false) {
    args = mapArguments(node.params, meta);
  }

  if (hasExpansion === true) {
    setupStatements = setupStatements.concat(setupStatements, mapArgumentsWithExpansion(node.params, meta));
  }

  // In coffeescript you can immediately assign an argument to a
  // member of `this`. Which looks like this: fn = (@a = 'A') ->
  // For our compilation we translate it like
  // fn = function() { this.a = arguments[0]; } as there is no 1 to 1
  // solution here
  setupStatements = setupStatements.concat(extractAssignStatementsByArguments(args, meta));

  var block = addReturnStatementToBlock(mapBlockStatement(node.body, meta), meta);

  block.body = setupStatements.concat(block.body);
  args = normalizeArguments(args, meta);

  if (node.bound === true) {
    return _astTypes.builders.arrowFunctionExpression(args, block);
  }

  return _astTypes.builders.functionExpression(null, args, block);
}

function getStatement(node) {
  if (_astTypes.namedTypes.Statement.check(node.value) !== true && node.parent) {
    return getStatement(node.parent);
  }
  return node;
}

function inParentScope(path, filter) {
  if (typeof filter !== 'function') {
    throw new Error('filter argument must be function');
  }
  var statement = getStatement(path);
  var scope = (0, _jscodeshift2.default)(path).closestScope().paths()[0];
  var statementsInScope = undefined;

  if (_astTypes.namedTypes.Program.check(scope.value)) {
    statementsInScope = scope.value.body;
  } else if (_astTypes.namedTypes.FunctionExpression.check(scope.value) || _astTypes.namedTypes.ArrowFunctionExpression.check(scope.value) || _astTypes.namedTypes.CatchClause.check(scope.value)) {
    statementsInScope = scope.value.body.body;
  } else {
    throw new Error('Can\'t recognize scope container of type ' + scope.value.type);
  }

  var indexInScope = (0, _findIndex2.default)(statementsInScope, function (node) {
    return statement.value === node;
  });

  var statements = statementsInScope.slice(0, indexInScope).filter(filter);
  var scopeStatement = getStatement(scope);

  if (_astTypes.namedTypes.Program.check(scope.value)) {
    return statements;
  }
  return inParentScope(scope.parent, filter).concat([scopeStatement].filter(function (_s) {
    return _s !== null;
  }).map(function (_s) {
    return _s.value;
  }).filter(filter)).concat(statements);
}

function insertSuperCall(path) {
  var classMethods = (0, _get2.default)(path, 'value.body.body') || [];
  var constructorIndex = (0, _findIndex2.default)(classMethods, { kind: 'constructor' });
  if (constructorIndex > -1) {
    var superCalls = (0, _jscodeshift2.default)(classMethods[constructorIndex]).find(_jscodeshift2.default.CallExpression, { callee: { name: 'super' } }).nodes();
    if (superCalls.length < 1) {
      classMethods[constructorIndex].value.body.body.unshift(_astTypes.builders.expressionStatement(_astTypes.builders.callExpression(_astTypes.builders.identifier('super'), [])));
    }
  }
}

function insertSuperCalls(ast) {
  (0, _jscodeshift2.default)(ast).find(_jscodeshift2.default.ClassDeclaration, function (path) {
    return (0, _get2.default)(path, 'superClass');
  }).forEach(insertSuperCall);

  (0, _jscodeshift2.default)(ast).find(_jscodeshift2.default.ClassExpression, function (path) {
    return (0, _get2.default)(path, 'superClass');
  }).forEach(insertSuperCall);

  return ast;
}

function insertVariableDeclarations(ast) {
  (0, _jscodeshift2.default)(ast).find(_jscodeshift2.default.AssignmentExpression, function (node) {
    return _astTypes.namedTypes.MemberExpression.check(node.left) !== true;
  }).filter(function (path) {
    var needle = { type: 'Identifier', name: path.value.left.name };

    var catchClauseParam = (0, _get2.default)(path, 'parent.parent.parent.value.param');

    if ((0, _get2.default)(path, 'parent.parent.parent.value.type') === 'CatchClause' && (0, _get2.default)(catchClauseParam, 'name') === needle.name) {
      return false;
    }

    var shadowedVariables = inParentScope(path, function (node) {
      if (_astTypes.namedTypes.VariableDeclaration.check(node)) {
        return (0, _findIndex2.default)(node.declarations, { id: needle }) > -1;
      }

      if (_astTypes.namedTypes.MethodDefinition.check(node) && node.value && (0, _findIndex2.default)(node.value, { left: needle })) {
        return true;
      }

      if (_astTypes.namedTypes.ExpressionStatement.check(node)) {
        if ((_astTypes.namedTypes.FunctionExpression.check(node.expression) || _astTypes.namedTypes.ArrowFunctionExpression.check(node.expression)) && (0, _findIndex2.default)(node.expression.params, needle) > -1) {
          return true;
        }

        if (_astTypes.namedTypes.AssignmentExpression.check(node.expression)) {
          if (node.expression.left.name === path.value.left.name) {
            return true;
          } else if (_astTypes.namedTypes.FunctionExpression.check(node.expression.right) || _astTypes.namedTypes.ArrowFunctionExpression.check(node.expression.right)) {
            if ((0, _findIndex2.default)(node.expression.right.params, needle) > -1) {
              return true;
            } else if ((0, _findIndex2.default)(node.expression.right.params, { type: 'AssignmentExpression', left: { name: path.value.left.name } }) > -1) {
              return true;
            }
          }
        }
      }

      if (_astTypes.namedTypes.ReturnStatement.check(node)) {
        if (_astTypes.namedTypes.AssignmentExpression.check(node.argument)) {
          return node.argument.left.name === path.value.left.name;
        }
      }
      return false;
    });

    return shadowedVariables.length < 1;
  }).forEach(function (path) {
    if (path.parent && path.parent.value.type === 'ExpressionStatement') {
      (0, _jscodeshift2.default)(path).replaceWith(function (_path) {
        return _astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(_path.value.left, _path.value.right)]);
      });
    } else {
      var body = (0, _jscodeshift2.default)(path).closestScope().nodes()[0].body;
      if (body.body !== undefined) {
        body = body.body;
      }
      body.unshift(_astTypes.builders.variableDeclaration('var', [_astTypes.builders.variableDeclarator(path.value.left, null)]));
    }
  });

  return ast;
}

function mapSwitchCases(cases, meta) {
  return cases.reduce(function (arr, nodes) {
    var tests = (0, _isArray2.default)(nodes[0]) ? nodes[0] : [nodes[0]];
    var switchCases = tests.filter(function (expr) {
      return !!expr;
    }).map(function (expr) {
      return _astTypes.builders.switchCase(mapExpression(expr), []);
    });
    var consequent = mapBlockStatements((0, _get2.default)(nodes, 1), meta);

    if (switchCases.length > 0) {
      switchCases[switchCases.length - 1].consequent = consequent;
    } else if (switchCases.length === 0 && !!nodes[1]) {
      switchCases.push(_astTypes.builders.switchCase(null, consequent));
    }

    return arr.concat((0, _flatten2.default)(switchCases));
  }, []);
}

function mapSwitchStatement(node, meta) {
  var cases = [];

  if (node.cases && node.cases.length > 0) {
    cases = cases.concat(node.cases);
  }

  if (node.otherwise) {
    cases.push([null, node.otherwise]);
  }

  return _astTypes.builders.switchStatement(mapExpression(node.subject, meta), mapSwitchCases(cases, meta));
}

function mapForStatement(node, meta) {
  if (node.object === false) {
    return _astTypes.builders.forInStatement(_astTypes.builders.variableDeclaration('let', [_astTypes.builders.variableDeclarator(mapExpression(node.name, _extends({}, meta, { left: true })), null)]), mapExpression(node.source, meta), mapBlockStatement(node.body, meta));
  } else if (node.object === true) {
    var args = _astTypes.builders.arrayPattern(pluck(node, ['index', 'name']).filter(function (expr) {
      return !!expr;
    }).map(function (expr) {
      return mapExpression(expr, meta);
    }));
    return _astTypes.builders.forInStatement(_astTypes.builders.variableDeclaration('let', [_astTypes.builders.variableDeclarator(args, null)]), _astTypes.builders.memberExpression(_astTypes.builders.identifier('Object'), _astTypes.builders.callExpression(_astTypes.builders.identifier('entries'), [mapExpression(node.source, meta)])), mapBlockStatement(node.body, meta));
  }
}

function mapLeftHandForExpression(node, meta) {
  if (node.step !== undefined) {
    return _astTypes.builders.memberExpression(mapExpression(node.source, meta), _astTypes.builders.callExpression(_astTypes.builders.identifier('filter'), [_astTypes.builders.arrowFunctionExpression([_astTypes.builders.identifier('_'), _astTypes.builders.identifier('_i')], _astTypes.builders.blockStatement([_astTypes.builders.returnStatement(_astTypes.builders.logicalExpression('||', _astTypes.builders.binaryExpression('===', _astTypes.builders.identifier('_i'), _astTypes.builders.literal(0)), _astTypes.builders.binaryExpression('===', _astTypes.builders.binaryExpression('%', _astTypes.builders.identifier('_i'), _astTypes.builders.binaryExpression('+', mapExpression(node.step, meta), _astTypes.builders.literal(1))), _astTypes.builders.literal(0))))], _recast2.default.parse('return _i === 0 || _i % (2 + 1) == 0;').program.body))]));
  }

  return mapExpression(node.source, meta);
}

function mapForExpression(node, meta) {
  var leftHand = mapLeftHandForExpression(node, meta);
  return _astTypes.builders.memberExpression(leftHand, _astTypes.builders.callExpression(_astTypes.builders.identifier('map'), [_astTypes.builders.arrowFunctionExpression([mapExpression(node.name, meta)], addReturnStatementToBlock(mapBlockStatement(node.body, meta)))]));
}

function mapSplatParam(node, meta) {
  return _astTypes.builders.restElement(mapExpression(node, meta));
}

function mapParam(node, meta) {
  if (node.value !== undefined && node.value !== null) {
    return mapExpression(mapParamToAssignment(node), meta);
  } else if (node.splat === true) {
    return mapSplatParam(node.name, meta);
  }
  return mapExpression(node.name, meta);
}

function mapSplat(node, meta) {
  return _astTypes.builders.spreadElement(mapExpression(node.name, meta));
}

function addReturnStatementsToSwitch(node) {
  node.cases = node.cases.map(function (switchCase) {
    switchCase.consequent = lastReturnStatement(switchCase.consequent);
    return switchCase;
  });
  return node;
}

function addBreakStatementsToSwitch(node) {
  node.cases = node.cases.map(function (switchCase) {
    if (switchCase.test !== null) {
      switchCase.consequent = lastBreakStatement(switchCase.consequent);
    }
    return switchCase;
  });
  return node;
}

function mapSwitchExpression(node, meta) {
  return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([addReturnStatementsToSwitch(mapSwitchStatement(node, meta))])), []);
}

function fallback(node, meta) {
  var compiled = node.compile(meta);
  return _recast2.default.parse(_recast2.default.prettyPrint(_recast2.default.parse(compiled), meta.options)).program.body[0].expression;
}

function mapExistentialExpression(node, meta) {
  return fallback(node, meta);
}

function mapNewExpression(node, meta) {
  var constructor = node.first || node.variable;
  var args = node.args ? mapArguments(node.args, meta) : [];
  return _astTypes.builders.newExpression(mapExpression(constructor, meta), args);
}

function conditionalStatementAsExpression(node, meta) {
  var conditionalStatement = mapConditionalStatement(node, meta);

  if (conditionalStatement.type === 'IfStatement') {
    return _astTypes.builders.callExpression(_astTypes.builders.arrowFunctionExpression([], _astTypes.builders.blockStatement([addReturnStatementToIfBlocks(conditionalStatement)])), []);
  }

  return conditionalStatement.expression;
}

// function mapComment(node) {
//   const comment = b.block(node.comment);
//   return comment;
// }

function mapWhileLoop(node, meta) {
  return _astTypes.builders.whileStatement(mapExpression(node.condition, meta), mapBlockStatement(node.body, meta));
}

function mapThrowStatement(node, meta) {
  return _astTypes.builders.throwStatement(mapExpression(node.expression, meta));
}

function mapWhileExpression(node, meta) {
  return fallback(node, meta);
}

function mapExpression(node, meta) {
  var type = node.constructor.name;

  if (node.properties && node.properties.length > 0) {
    return mapMemberExpression(node, meta);
  } else if (type === 'Range') {
    return mapRange(node, meta);
  } else if (type === 'If') {
    return conditionalStatementAsExpression(node, meta);
  } else if (type === 'Parens' && (0, _get2.default)(node, 'body.expressions[0]')) {
    return _astTypes.builders.parenthesizedExpression(mapExpression((0, _get2.default)(node, 'body.expressions[0]'), meta));
  } else if (type === 'Parens') {
    return _astTypes.builders.parenthesizedExpression(mapExpression(node.body, meta));
  } else if (type === 'Arr') {
    return mapValue({ base: node }, meta);
  } else if (type === 'Try') {
    return mapTryExpression(node, meta);
  } else if (type === 'Call' && node.isNew === true) {
    return mapNewExpression(node, meta);
  } else if (type === 'Op' && node.operator === 'new') {
    return mapNewExpression(node, meta);
  } else if (type === 'Existence') {
    return mapExistentialExpression(node, meta);
  } else if (type === 'While') {
    return mapWhileExpression(node, meta);
  } else if (type === 'Switch') {
    return mapSwitchExpression(node, meta);
  } else if (type === 'Splat') {
    return mapSplat(node, meta);
  } else if (type === 'Assign') {
    return mapAssignmentExpression(node, meta);
  } else if (type === 'Slice') {
    return mapSlice(node, meta);
  } else if (type === 'For') {
    return mapForExpression(node, meta);
  } else if (type === 'Param') {
    return mapParam(node, meta);
  } else if (type === 'Class') {
    return mapClassExpression(node, meta);
  } else if (type === 'Extends') {
    return mapExpression(node.parent, meta);
  } else if (type === 'Code') {
    // Code is just a stupid word for function
    return mapFunction(node, meta);
  } else if (type === 'Index') {
    return mapExpression(node.index, meta);
  } else if (type === 'Access') {
    return mapExpression(node.name, meta);
  } else if (type === 'Literal') {
    return mapLiteral({ base: node }, meta);
  } else if (type === 'In') {
    return mapInArrayExpression(node, meta);
  } else if (type === 'Value') {
    return mapValue(node, meta);
  } else if (type === 'Op') {
    return mapOp(node, meta);
  } else if (type === 'Call') {
    return mapCall(node, meta);
  }

  throw new Error('can\'t convert node of type: ' + type + ' to Expression - not recognized');
}

function mapParamToAssignment(node) {
  var assignment = {
    variable: node.name,
    value: node.value
  };
  assignment.constructor = { name: 'Assign' };
  return assignment;
}

function mapAssignmentExpression(node, meta) {
  var variable = undefined;
  if ((0, _get2.default)(node, 'variable.base.properties.length') > 0) {
    variable = mapAssignmentPattern(node.variable.base, meta);
  } else {
    variable = mapExpression(node.variable, meta);
  }

  var assignment = _astTypes.builders.assignmentExpression('=', variable, mapExpression(node.value, meta));

  if (node.context === '||=') {
    return _astTypes.builders.logicalExpression('||', variable, assignment);
  }

  if (node.context === '?=') {
    return _astTypes.builders.conditionalExpression(_astTypes.builders.binaryExpression('!=', variable, _astTypes.builders.identifier('null')), variable, assignment);
  }

  return assignment;
}

function mapObjectPatternItem(node, meta) {
  var type = node.constructor.name;
  if (type === 'Value') {
    return mapLiteral(node, meta);
  } else if (type === 'Assign') {
    if (node.value.base.properties) {
      return mapObjectPattern(node.value.base.properties, meta);
    }
    return mapExpression(node.value, meta);
  }

  throw new Error('can\'t convert node of type: ' + type + ' to ObjectPatternItem - not recognized');
}

function mapObjectPattern(nodes, meta) {
  return _astTypes.builders.objectPattern(nodes.map(function (node) {
    var operatorToken = node.operatorToken;

    var prop = undefined;
    prop = _astTypes.builders.property('init', mapKey(node.variable || node, meta), mapObjectPatternItem(node, meta));

    if (operatorToken === undefined) {
      prop.shorthand = true;
    }
    return prop;
  }));
}

function mapArrayPattern(node, meta) {
  return _astTypes.builders.arrayPattern(node.objects.map(function (prop) {
    var type = prop.base.constructor.name;
    if (type === 'Literal') {
      return mapLiteral(prop, meta);
    } else if (type === 'Arr') {
      return mapArrayPattern(prop.base, meta);
    } else if (type === 'Obj') {
      return mapObjectPattern(prop.base.properties, _extends({}, meta, { left: true }));
    }
  }));
}

function mapAssignmentPattern(node, meta) {
  // that's a destructuring assignment
  var type = node.constructor.name;

  if (type === 'Obj' && node.properties) {
    return mapObjectPattern(node.properties, meta);
  } else if (type === 'Arr') {
    return mapArrayPattern(node, meta);
  }

  return mapExpression(node, meta);
}

function coffeeParse(source) {
  var ast = (0, _coffeeScript.nodes)(source);
  return ast;
}

function transpile(ast, meta) {
  if (meta === undefined) {
    meta = {};
  }

  if (!meta.scope) {
    meta.scope = new _scope.Scope(null, coffeeParse, null, []);
    meta.indent = ' ';
  }

  var program = mapBlockStatement(ast, meta, _astTypes.builders.program);
  return program;
}

function compile(source, opts) {
  var doubleSemicolon = /\;+/g;
  opts = opts || { tabWidth: 2, quote: 'double' };

  var _compile = (0, _compose2.default)(
  // hack because of double semicolon
  function (compiledSource) {
    return _extends({}, compiledSource, { code: compiledSource.code.replace(doubleSemicolon, ';') });
  }, function (jsAst) {
    return _recast2.default.print(jsAst, opts);
  }, insertSuperCalls, insertVariableDeclarations, function (csAst) {
    return transpile(csAst, { options: opts });
  }, coffeeParse);

  return _compile(source).code;
}
